<?php
/**
 * Order Status Rules for WooCommerce - Core Class
 *
 * @version 1.8.0
 * @since   1.0.0
 *
 * @author  Algoritmika Ltd.
 */

if ( ! defined( 'ABSPATH' ) ) exit; // Exit if accessed directly

if ( ! class_exists( 'Alg_WC_Order_Status_Rules_Core' ) ) :

class Alg_WC_Order_Status_Rules_Core {

	/**
	 * Constructor.
	 *
	 * @version 1.8.0
	 * @since   1.0.0
	 *
	 * @todo    [next] [!] (dev) process zero triggers differently, i.e. directly, without crons
	 * @todo    [next] (dev) `save_status_change`: maybe set priority to e.g. `1` or `0`?
	 * @todo    [next] (dev) remove `alg_wc_order_status_rules_plugin_enabled` (or move `process_rules_manual`, etc. inside the `alg_wc_order_status_rules_plugin_enabled`)
	 */
	function __construct() {
		if ( 'yes' === get_option( 'alg_wc_order_status_rules_plugin_enabled', 'yes' ) ) {
			// Track order status change
			add_action( 'woocommerce_order_status_changed', array( $this, 'save_status_change' ), 10, 4 );
			// Process rules via URL
			if ( 'yes' === get_option( 'alg_wc_order_status_rules_allow_url', 'no' ) ) {
				add_action( 'init', array( $this, 'process_rules_url' ), PHP_INT_MAX );
			}
			// Admin order edit page meta box
			require_once( 'class-alg-wc-order-status-rules-meta-boxes.php' );
			// My account > Orders
			require_once( 'class-alg-wc-order-status-rules-my-account.php' );
			// Compatibility
			require_once( 'class-alg-wc-order-status-rules-compatibility.php' );
		}
		// Crons
		$this->crons = require_once( 'class-alg-wc-order-status-rules-crons.php' );
		// Process rules manually
		add_action( 'alg_wc_order_status_after_save_settings', array( $this, 'process_rules_manual' ) );
	}

	/**
	 * add_to_log.
	 *
	 * @version 1.4.0
	 * @since   1.4.0
	 */
	function add_to_log( $message ) {
		if ( function_exists( 'wc_get_logger' ) && ( $log = wc_get_logger() ) ) {
			$log->log( 'info', $message, array( 'source' => 'order-status-rules' ) );
		}
	}

	/**
	 * do_debug.
	 *
	 * @version 1.4.0
	 * @since   1.4.0
	 */
	function do_debug() {
		if ( ! isset( $this->do_debug ) ) {
			$this->do_debug = ( 'yes' === get_option( 'alg_wc_order_status_rules_debug', 'no' ) );
		}
		return $this->do_debug;
	}

	/**
	 * on_no_history.
	 *
	 * @version 1.8.0
	 * @since   1.4.0
	 */
	function on_no_history() {
		if ( ! isset( $this->on_no_history ) ) {
			$this->on_no_history = get_option( 'alg_wc_order_status_rules_no_history', 'use_date_modified' );
		}
		return $this->on_no_history;
	}

	/**
	 * process_rules_url.
	 *
	 * @version 1.3.0
	 * @since   1.3.0
	 *
	 * @todo    [maybe] (dev) optional "key" (for security)
	 * @todo    [maybe] (dev) optional "rule ID to process"
	 */
	function process_rules_url() {
		if ( isset( $_REQUEST['alg_wc_order_status_rules_process_rules'] ) ) {
			$this->process_rules();
		}
	}

	/**
	 * process_rules_manual.
	 *
	 * @version 1.2.0
	 * @since   1.2.0
	 */
	function process_rules_manual() {
		if ( 'yes' === get_option( 'alg_wc_order_status_rules_run_now', 'no' ) ) {
			update_option( 'alg_wc_order_status_rules_run_now', 'no' );
			$this->process_rules( false );
			if ( method_exists( 'WC_Admin_Settings', 'add_message' ) ) {
				WC_Admin_Settings::add_message( __( '"Run all rules now" tool executed.', 'order-status-rules-for-woocommerce' ) );
			} else {
				add_action( 'admin_notices', array( $this, 'admin_notice_process_rules_manual' ) );
			}
		}
	}

	/**
	 * admin_notice_process_rules_manual.
	 *
	 * @version 1.2.0
	 * @since   1.2.0
	 */
	function admin_notice_process_rules_manual() {
		echo '<div class="notice notice-warning is-dismissible"><p><strong>' .
			__( '"Run all rules now" tool executed.', 'order-status-rules-for-woocommerce' ) . '</strong></p></div>';
	}

	/**
	 * get_time_remaining.
	 *
	 * @version 1.5.0
	 * @since   1.0.1
	 *
	 * @todo    [next] (test) $skip_days (especially for `unit != hour`)
	 * @todo    [next] (dev) simplify $skip_days
	 */
	function get_time_remaining( $last_record_time, $trigger_time, $skip_days = false, $step = 3600 ) {
		$current_time = current_time( 'timestamp' );
		$time_passed  = ( $current_time - $last_record_time );
		if ( ! empty( $skip_days ) && count( $skip_days ) < 7 ) {
			$start_time    = $last_record_time + 1;
			$counter_total = $start_time;
			$counter_ok    = 0;
			while ( $counter_ok <= $trigger_time ) {
				$day_of_the_week = date( 'N', $counter_total );
				if ( ! in_array( $day_of_the_week, $skip_days ) ) {
					$counter_ok += $step;
				}
				$counter_total += $step;
			}
			$trigger_time = ( $counter_total - $start_time - $step );
		}
		return ( $trigger_time - $time_passed );
	}

	/**
	 * get_order_status_change_history.
	 *
	 * @version 1.7.0
	 * @since   1.4.0
	 *
	 * @todo    [next] (dev) use `getTimestamp()`, not `getOffsetTimestamp()`
	 */
	function get_order_status_change_history( $order_id ) {
		$data = get_post_meta( $order_id, '_alg_wc_order_status_change_history', true );
		if ( empty( $data ) && 'do_nothing' != $this->on_no_history() ) {
			$order = wc_get_order( $order_id );
			if ( ( $date = ( 'use_date_created' === $this->on_no_history() ? $order->get_date_created() : $order->get_date_modified() ) ) ) {
				$data = array(
					array(
						'time' => $date->getOffsetTimestamp(),
						'from' => 'N/A',
						'to'   => $order->get_status(),
					),
				);
			}
		}
		return $data;
	}

	/**
	 * check_order_items.
	 *
	 * @version 1.6.0
	 * @since   1.6.0
	 */
	function check_order_items( $order, $values, $type ) {
		foreach ( $order->get_items() as $item ) {
			switch ( $type ) {
				case 'product':
					if ( in_array( $item['product_id'], $values ) || in_array( $item['variation_id'], $values ) ) {
						return true;
					}
					break;
				case 'product_cat':
				case 'product_tag':
					if ( has_term( $values, $type, $item['product_id'] ) ) {
						return true;
					}
					break;
			}
		}
		return false;
	}

	/**
	 * init_options.
	 *
	 * @version 1.6.0
	 * @since   1.6.0
	 *
	 * @todo    [next] (dev) `$this->options`: rename?
	 * @todo    [next] (dev) `$this->options['from']`: redo?
	 */
	function init_options() {
		if ( ! isset( $this->options ) ) {
			$this->options = array(
				'enabled'            => get_option( 'alg_wc_order_status_rules_enabled',           array() ),
				'from'               => get_option( 'alg_wc_order_status_rules_from',              array() ),
				'to'                 => get_option( 'alg_wc_order_status_rules_to',                array() ),
				'time_triggers'      => get_option( 'alg_wc_order_status_rules_time_trigger',      array() ),
				'time_trigger_units' => get_option( 'alg_wc_order_status_rules_time_trigger_unit', array() ),
				'skip_days'          => get_option( 'alg_wc_order_status_rules_skip_days',         array() ),
				'gateways'           => get_option( 'alg_wc_order_status_rules_gateways',          array() ),
				'products'           => get_option( 'alg_wc_order_status_rules_products',          array() ),
				'product_cats'       => get_option( 'alg_wc_order_status_rules_product_cats',      array() ),
				'product_tags'       => get_option( 'alg_wc_order_status_rules_product_tags',      array() ),
				'titles'             => get_option( 'alg_wc_order_status_rules_title',             array() ),
			);
			$this->options['from'] = array_slice( $this->options['from'], 0, apply_filters( 'alg_wc_order_status_rules_rules_total', 1 ), true );
		}
	}

	/**
	 * do_apply_rule.
	 *
	 * @version 1.8.0
	 * @since   1.6.0
	 */
	function do_apply_rule( $i, $args ) {
		$this->init_options();
		return (
			( isset( $this->options['enabled'][ $i ] ) && 'yes' === $this->options['enabled'][ $i ] ) &&
			( 'wc-' . $args['order_status'] == $args['from'] && 'wc-' . $args['last_record']['to'] == $args['from'] ) &&
			( empty( $this->options['gateways'][ $i ] )     || in_array( $args['gateway'], $this->options['gateways'][ $i ] ) ) &&
			( empty( $this->options['products'][ $i ] )     || $this->check_order_items( $args['order'], $this->options['products'][ $i ],     'product' ) ) &&
			( empty( $this->options['product_cats'][ $i ] ) || $this->check_order_items( $args['order'], $this->options['product_cats'][ $i ], 'product_cat' ) ) &&
			( empty( $this->options['product_tags'][ $i ] ) || $this->check_order_items( $args['order'], $this->options['product_tags'][ $i ], 'product_tag' ) ) &&
			( isset( $this->options['time_triggers'][ $i ] ) ) &&
			( isset( $this->options['to'][ $i ] ) )
		);
	}

	/**
	 * process_rules.
	 *
	 * @version 1.7.0
	 * @since   1.0.0
	 *
	 * @todo    [next] (dev) code refactoring: `args`, `step`, `skip`?
	 * @todo    [next] (dev) optimization: `$this->options['from']`: enabled rules only
	 * @todo    [next] (dev) optimization: meta query: `_alg_wc_order_status_change_history` not empty? (only if `'do_nothing' === $this->on_no_history()`)
	 * @todo    [next] (dev) log: add more info
	 * @todo    [next] (desc) use `alg_wc_order_status_rules_process_rules_time_run`
	 */
	function process_rules( $do_die = true ) {
		update_option( 'alg_wc_order_status_rules_process_rules_time_run', time() );
		if ( $this->do_debug() ) {
			$this->add_to_log( __( 'Process rules: Started', 'order-status-rules-for-woocommerce' ) .
				( ! $do_die ? ' (' . __( 'manual', 'order-status-rules-for-woocommerce' ) . ')' : '' ) );
		}
		$this->init_options();
		if ( ! empty( $this->options['from'] ) ) {
			$orders = wc_get_orders( array(
				'limit'    => -1,
				'status'   => $this->options['from'],
				'return'   => 'ids',
			) );
			foreach ( $orders as $order_id ) {
				if ( $this->do_debug() ) {
					$this->add_to_log( sprintf( __( 'Process rules: Order #%s', 'order-status-rules-for-woocommerce' ), $order_id ) );
				}
				$status_history = $this->get_order_status_change_history( $order_id );
				if ( ! empty( $status_history ) ) {
					$status_history = array_reverse( $status_history, true );
					$last_record    = current( $status_history );
					$order          = wc_get_order( $order_id );
					$order_status   = $order->get_status();
					$gateway        = $order->get_payment_method();
					foreach ( $this->options['from'] as $i => $from ) {
						$step = $this->get_trigger_unit_step( ( isset( $this->options['time_trigger_units'][ $i ] ) ? $this->options['time_trigger_units'][ $i ] : 'hour' ) );
						$skip = ( isset( $this->options['skip_days'][ $i ] ) ? $this->options['skip_days'][ $i ] : false );
						$args = array(
							'order_status' => $order_status,
							'from'         => $from,
							'last_record'  => $last_record,
							'gateway'      => $gateway,
							'order'        => $order,
						);
						if (
							$this->do_apply_rule( $i, $args ) &&
							( $this->get_time_remaining( $last_record['time'], $this->options['time_triggers'][ $i ] * $step, $skip, $step ) <= 0 )
						) {
							$rule = sprintf( __( 'Rule #%s', 'order-status-rules-for-woocommerce' ), $i ) .
								( ! empty( $this->options['titles'][ $i ] ) ? ': ' . $this->options['titles'][ $i ] : '' );
							$note = sprintf( __( 'Status updated by "Order Status Rules for WooCommerce" plugin (%s).', 'order-status-rules-for-woocommerce' ), $rule );
							$order->update_status( substr( $this->options['to'][ $i ], 3 ), $note );
							if ( $this->do_debug() ) {
								$this->add_to_log( sprintf( __( 'Process rules: Order #%s: from %s to %s (%s)', 'order-status-rules-for-woocommerce' ),
									$order_id, $from, $this->options['to'][ $i ], $rule ) );
							}
							break;
						}
					}
				} elseif ( $this->do_debug() ) {
					$this->add_to_log( sprintf( __( 'Process rules: Order #%s: No order status change history found', 'order-status-rules-for-woocommerce' ), $order_id ) );
				}
			}
		}
		if ( $this->do_debug() ) {
			$this->add_to_log( __( 'Process rules: Finished', 'order-status-rules-for-woocommerce' ) );
		}
		if ( $do_die ) {
			die();
		}
	}

	/**
	 * get_trigger_unit_step.
	 *
	 * @version 1.4.0
	 * @since   1.4.0
	 */
	function get_trigger_unit_step( $unit ) {
		switch ( $unit ) {
			case 'second':
				return 1;
			case 'minute':
				return MINUTE_IN_SECONDS;
			case 'hour':
				return HOUR_IN_SECONDS;
			case 'day':
				return DAY_IN_SECONDS;
			case 'week':
				return WEEK_IN_SECONDS;
			default: // 'hour'
				return HOUR_IN_SECONDS;
		}
	}

	/**
	 * save_status_change.
	 *
	 * @version 1.0.0
	 * @since   1.0.0
	 *
	 * @todo    [next] (dev) when `$from` doesn't exist `woocommerce_order_status_changed` is not called; check `do_action( 'woocommerce_order_status_' . $status_transition['to'], $this->get_id(), $this );`
	 * @todo    [next] (dev) save `time()`, not `current_time()`
	 * @todo    [maybe] (dev) mark status change as "changed by plugin" (vs "changed manually/otherwise")
	 */
	function save_status_change( $order_id, $from, $to, $order ) {
		$status_history = get_post_meta( $order_id, '_alg_wc_order_status_change_history', true );
		if ( empty( $status_history ) ) {
			$status_history = array();
		}
		$status_history[] = array(
			'time' => current_time( 'timestamp' ),
			'from' => $from,
			'to'   => $to,
		);
		update_post_meta( $order_id, '_alg_wc_order_status_change_history', $status_history );
	}

	/**
	 * get_date_time_format.
	 *
	 * @version 1.4.0
	 * @since   1.4.0
	 */
	function get_date_time_format() {
		return get_option( 'date_format' ) . ' ' . get_option( 'time_format' );
	}

}

endif;

return new Alg_WC_Order_Status_Rules_Core();
